from argparse import ArgumentParser

from rosy.types import DomainId
from rosy.utils import DEFAULT_DOMAIN_ID, get_domain_id


def get_node_arg_parser(
    default_node_name: str = None,
    default_domain_id: DomainId = DEFAULT_DOMAIN_ID,
    **kwargs,
) -> ArgumentParser:
    """
    Returns an argument parser with the following node-specific arguments:
    - `--name`
    - `--domain-id`

    Args:
        default_node_name:
            Default node name. If not given, the argument `--name` is required.
        default_domain_id:
            Default Domain ID.
        **kwargs:
            These are passed to the `ArgumentParser` constructor.
    """

    parser = ArgumentParser(**kwargs)
    add_node_args(parser, default_node_name, default_domain_id)
    return parser


def add_node_args(
    parser: ArgumentParser,
    default_node_name: str = None,
    default_domain_id: DomainId = DEFAULT_DOMAIN_ID,
) -> None:
    """
    Adds node-specific arguments to an argument parser:
    - `--name`
    - `--domain-id`

    Args:
        parser:
            Argument parser to which the arguments will be added.
        default_node_name:
            Default node name. If not given, the argument is required.
        default_domain_id:
            Default Domain ID.
    """

    add_node_name_arg(parser, default_node_name)
    add_domain_id_arg(parser, default_domain_id)


def add_node_name_arg(parser: ArgumentParser, default: str = None) -> None:
    """
    Adds a `--name` argument to an argument parser.

    Args:
        parser:
            Argument parser to which the argument will be added.
        default:
            Default node name. If not given, the argument will be required.
    """

    arg_args = (
        dict(
            default=default,
            help="Node name. Default: %(default)s",
        )
        if default is not None
        else dict(
            required=True,
            help="Node name.",
        )
    )

    parser.add_argument("--name", **arg_args)


def add_domain_id_arg(
    parser: ArgumentParser,
    default: DomainId = DEFAULT_DOMAIN_ID,
) -> None:
    """
    Adds a `--domain-id` argument to an argument parser.

    Args:
        parser:
            Argument parser to which the argument will be added.
        default:
            Default Domain ID to use if not given and the `ROSY_DOMAIN_ID`
            environment variable is not set.
    """

    parser.add_argument(
        "--domain-id",
        default=get_domain_id(default),
        help="""The Domain ID allows multiple rosy meshes to coexist on the
        same network. Only nodes with the same domain ID can connect to each
        other. This can also be set with the `ROSY_DOMAIN_ID` environment
        variable.
        Default: %(default)r""",
    )



import logging
from asyncio import Lock, open_connection, open_unix_connection
from collections.abc import Iterable
from typing import NamedTuple

from rosy.asyncio import LockableWriter, Reader, Writer, close_ignoring_errors
from rosy.network import get_hostname
from rosy.specs import (
    ConnectionSpec,
    IpConnectionSpec,
    MeshNodeSpec,
    NodeId,
    UnixConnectionSpec,
)
from rosy.types import Host

logger = logging.getLogger(__name__)


class PeerConnection(NamedTuple):
    reader: Reader
    writer: LockableWriter

    async def close(self) -> None:
        await close_ignoring_errors(self.writer)

    def is_closing(self) -> bool:
        return self.writer.is_closing()


class PeerConnectionBuilder:
    def __init__(self, host: Host = None):
        self.host = host or get_hostname()

    async def build(
        self, conn_specs: Iterable[ConnectionSpec]
    ) -> tuple[Reader, Writer]:
        reader_writer = None
        for conn_spec in conn_specs:
            try:
                reader_writer = await self._get_connection(conn_spec)
            except (ConnectionError, IOError) as e:
                logger.error(f"Error connecting to {conn_spec}: {e!r}")
                continue

            if reader_writer is not None:
                break

        if reader_writer is None:
            raise ConnectionError("Could not connect to any connection spec")

        return reader_writer

    async def _get_connection(
        self, conn_spec: ConnectionSpec
    ) -> tuple[Reader, Writer] | None:
        if isinstance(conn_spec, IpConnectionSpec):
            return await open_connection(
                host=conn_spec.host,
                port=conn_spec.port,
                family=conn_spec.family,
            )
        elif isinstance(conn_spec, UnixConnectionSpec):
            if conn_spec.host != self.host:
                return None

            return await open_unix_connection(path=conn_spec.path)
        else:
            raise ValueError(f"Unrecognized connection spec: {conn_spec}")


class PeerConnectionManager:
    def __init__(self, conn_builder: PeerConnectionBuilder):
        self.conn_builder = conn_builder

        self._connections: dict[NodeId, PeerConnection] = {}
        self._connections_lock = Lock()

    async def get_connection(self, node: MeshNodeSpec) -> PeerConnection:
        async with self._connections_lock:
            connection = await self._get_cached_connection(node)
            if connection:
                return connection

            logger.debug(f"Connecting to node: {node.id}")
            reader, writer = await self.conn_builder.build(node.connection_specs)

            writer = LockableWriter(writer)

            connection = PeerConnection(reader, writer)
            self._connections[node.id] = connection
            return connection

    async def _get_cached_connection(self, node: MeshNodeSpec) -> PeerConnection | None:
        connection = self._connections.get(node.id, None)
        if not connection:
            return None

        if connection.is_closing():
            await self.close_connection(node)
            return None

        return connection

    async def close_connection(self, node: MeshNodeSpec) -> None:
        connection = self._connections.pop(node.id, None)
        if connection:
            await connection.close()



import logging
from abc import ABC, abstractmethod
from typing import Awaitable, Callable

from rosy.specs import MeshNodeSpec, MeshTopologySpec
from rosy.utils import ALLOWED_EXCEPTIONS

TopologyChangedCallback = Callable[[MeshTopologySpec], Awaitable[None]]

logger = logging.getLogger(__name__)


class NodeDiscovery(ABC):
    topology_changed_callback: TopologyChangedCallback | None

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.stop()

    @property
    @abstractmethod
    def topology(self) -> MeshTopologySpec: ...

    @abstractmethod
    async def start(self) -> None: ...

    @abstractmethod
    async def stop(self) -> None: ...

    @abstractmethod
    async def register_node(self, node: MeshNodeSpec) -> None: ...

    @abstractmethod
    async def update_node(self, node: MeshNodeSpec) -> None: ...

    async def _call_topology_changed_callback(self) -> None:
        if self.topology_changed_callback is None:
            return

        try:
            await self.topology_changed_callback(self.topology)
        except ALLOWED_EXCEPTIONS:
            raise
        except Exception as e:
            logger.exception(
                "Error calling topology changed callback",
                exc_info=e,
            )



from rosy.node.loadbalancing import ServiceLoadBalancer, TopicLoadBalancer
from rosy.node.topology import MeshTopologyManager
from rosy.specs import MeshNodeSpec
from rosy.types import Service, Topic


class PeerSelector:
    def __init__(
        self,
        topology_manager: MeshTopologyManager,
        topic_load_balancer: TopicLoadBalancer,
        service_load_balancer: ServiceLoadBalancer,
    ):
        self.topology_manager = topology_manager
        self.topic_load_balancer = topic_load_balancer
        self.service_load_balancer = service_load_balancer

    def get_nodes_for_topic(self, topic: Topic) -> list[MeshNodeSpec]:
        peers = self.topology_manager.get_nodes_listening_to_topic(topic)
        return self.topic_load_balancer.choose_nodes(peers, topic)

    def get_node_for_service(self, service: Service) -> MeshNodeSpec | None:
        peers = self.topology_manager.get_nodes_providing_service(service)
        return self.service_load_balancer.choose_node(peers, service)



import asyncio
import gzip
import hashlib
import logging
import pickle
import socket
from abc import ABC, abstractmethod
from random import Random
from typing import cast

from zeroconf import ServiceBrowser, ServiceInfo, ServiceStateChange, Zeroconf

from rosy.asyncio import cancel_task
from rosy.discovery.base import NodeDiscovery, TopologyChangedCallback
from rosy.specs import MeshNodeSpec, MeshTopologySpec
from rosy.types import DomainId

DEFAULT_TTL: int = 30

logger = logging.getLogger(__name__)


class ZeroconfNodeDiscovery(NodeDiscovery):
    def __init__(
        self,
        domain_id: DomainId,
        topology_changed_callback: TopologyChangedCallback = None,
        zc: Zeroconf = None,
        node_spec_codec: "NodeSpecCodec" = None,
        ttl: int = DEFAULT_TTL,
        rng: Random = None,
    ) -> None:
        self._service_type = build_service_type(domain_id)
        self.topology_changed_callback = topology_changed_callback
        self._zc = zc or Zeroconf()
        self._node_spec_codec = node_spec_codec or GzipPickleNodeSpecCodec()
        self._ttl = ttl
        self._rng = rng or Random()

        self._service_name_to_node: dict[str, MeshNodeSpec] = {}
        self._topology_changed_caller_task: asyncio.Task | None = None
        self._browser: ServiceBrowser | None = None
        self._node_monitors: dict[str, asyncio.Task] = {}
        self._topology_changed = asyncio.Event()
        self._service_info_lock = asyncio.Lock()

    @property
    def topology(self) -> MeshTopologySpec:
        nodes = list(self._service_name_to_node.values())
        return MeshTopologySpec(nodes)

    async def start(self) -> None:
        self._topology_changed_caller_task = asyncio.create_task(
            self._monitor_for_topology_changes()
        )

        self._browser = ServiceBrowser(
            self._zc,
            self._service_type,
            handlers=[self._on_service_state_change],
        )

    async def stop(self, timeout: float | None = 10) -> None:
        exceptions = []

        try:
            await cancel_task(self._topology_changed_caller_task, timeout=timeout)
        except Exception as e:
            exceptions.append(e)

        try:
            await asyncio.wait_for(asyncio.to_thread(self._browser.cancel), timeout)
        except Exception as e:
            exceptions.append(e)

        tasks = list(self._node_monitors.values())
        for task in tasks:
            try:
                await cancel_task(task, timeout=timeout)
            except Exception as e:
                exceptions.append(e)

        try:
            await asyncio.wait_for(asyncio.to_thread(self._zc.close), timeout)
        except Exception as e:
            exceptions.append(e)

        if exceptions:
            # TODO Replace with ExceptionGroup when we drop Python 3.10
            raise Exception(
                "Exceptions occurred while stopping ZeroconfNodeDiscovery",
                exceptions,
            )

    async def register_node(self, node: MeshNodeSpec) -> None:
        logger.debug(f"Registering node: {node}")
        info = self._build_service_info(node)
        await (await self._zc.async_register_service(info))

    async def update_node(self, node: MeshNodeSpec) -> None:
        logger.debug(f"Updating node: {node}")
        info = self._build_service_info(node)
        await (await self._zc.async_update_service(info))

    def _build_service_info(self, node: MeshNodeSpec) -> ServiceInfo:
        return ServiceInfo(
            type_=self._service_type,
            name=f"{node.id.uuid}.{self._service_type}",
            port=0,
            properties=self._node_spec_codec.encode(node),
            server=get_mdns_fqdn(),
            host_ttl=self._ttl,
            other_ttl=self._ttl,
        )

    async def _monitor_for_topology_changes(self) -> None:
        while True:
            await self._topology_changed.wait()
            self._topology_changed.clear()
            await self._call_topology_changed_callback()

    def _on_service_state_change(
        self,
        zeroconf: Zeroconf,
        service_type: str,
        name: str,
        state_change: ServiceStateChange,
    ) -> None:
        """Called by the ServiceBrowser thread."""

        logger.debug(f"Service state change: {state_change.name} {name!r}")

        future = asyncio.run_coroutine_threadsafe(
            self._async_on_service_state_change(name, state_change),
            self._zc.loop,
        )
        future.result()

    async def _async_on_service_state_change(
        self,
        name: str,
        state_change: ServiceStateChange,
    ) -> None:
        if state_change in (ServiceStateChange.Added, ServiceStateChange.Updated):
            await self._start_monitoring_node(name)
        elif state_change == ServiceStateChange.Removed:
            await self._remove_node(name)

    async def _start_monitoring_node(self, name: str) -> None:
        """
        Why do we need to monitor the node? Why not just rely on the service
        state change events?

        Under normal circumstances, when a node exits, it will broadcast a
        notification, in which case a service state change event will be
        triggered immediately, and all is well.

        However, if a node dies unexpectedly and does not perform this cleanup
        step, or if there is a network failure between the nodes, `zeroconf`
        will not "pick this up" for a long time, like up to 18 minutes.
        Unfortunately, this is by design, and cannot be changed.

        To work around this, we periodically retrieve the node's service info,
        shortly after its TTL period. If the node is still alive, it should've
        already re-broadcast itself; otherwise, the TTL will have expired, and
        when we try to retrieve the service info again, it will be `None`,
        thus indicating that the node has left the mesh.
        """

        await self._stop_monitoring_node(name)
        self._node_monitors[name] = asyncio.create_task(self._monitor_node(name))

    async def _stop_monitoring_node(self, name: str) -> None:
        if name in self._node_monitors:
            await cancel_task(self._node_monitors[name])
            self._node_monitors.pop(name)

    async def _monitor_node(self, name: str) -> None:
        while await self._check_node(name):
            sleep_time = self._ttl + 1 + self._rng.random()
            await asyncio.sleep(sleep_time)

    async def _check_node(self, name: str) -> bool:
        self._check_this_task_is_current_monitor(name)

        logger.debug(f"Getting service info for: {name!r}")
        info = await asyncio.shield(self._get_service_info(name))

        if info is None:
            logger.error(f"Failed to get service info for: {name!r}")
            asyncio.create_task(self._remove_node(name))
            return False

        text = cast(bytes, info.text)
        node = self._node_spec_codec.decode(text)

        existing_node = self._service_name_to_node.get(name)
        if node != existing_node:
            if existing_node is None:
                logger.debug(f"Discovered node on mesh: {node}")
            else:
                logger.debug(f"Node updated: {node}")

            self._service_name_to_node[name] = node
            self._topology_changed.set()

        return True

    def _check_this_task_is_current_monitor(self, name: str) -> None:
        this_task = asyncio.current_task()
        monitor_task = self._node_monitors[name]
        if this_task is not monitor_task:
            raise RuntimeError(
                f"This task is not the current node monitor task. "
                f"this_task={this_task}; monitor_task={monitor_task}"
            )

    async def _get_service_info(self, name: str) -> ServiceInfo | None:
        async with self._service_info_lock:
            return await self._zc.async_get_service_info(self._service_type, name)

    async def _remove_node(self, name: str) -> None:
        await self._stop_monitoring_node(name)

        node = self._service_name_to_node.pop(name, None)
        if node is not None:
            logger.debug(f"Node left mesh: {node.id}")
            self._topology_changed.set()


def build_service_type(domain_id: DomainId) -> str:
    domain_id = hash_domain_id(domain_id)
    return f"_rosy-{domain_id}._tcp.local."


def hash_domain_id(domain_id: str, digest_size: int = 5) -> str:
    domain_id = domain_id.encode()
    domain_id = hashlib.blake2b(
        domain_id,
        digest_size=digest_size,
        usedforsecurity=False,
    )
    return domain_id.hexdigest()


def get_mdns_fqdn() -> str:
    fqdn = socket.getfqdn()

    if fqdn.endswith(".local."):
        return fqdn
    elif fqdn.endswith(".local"):
        return fqdn + "."
    elif fqdn.endswith("."):
        return fqdn + "local."
    else:
        return fqdn + ".local."


class NodeSpecCodec(ABC):
    @abstractmethod
    def encode(self, node: MeshNodeSpec) -> bytes: ...

    @abstractmethod
    def decode(self, data: bytes) -> MeshNodeSpec: ...


class GzipPickleNodeSpecCodec(NodeSpecCodec):
    """
    Not a lot of space in the UDP payload, so compress as much as possible.
    """

    def __init__(
        self,
        protocol: int = pickle.HIGHEST_PROTOCOL,
        compresslevel: int = 9,
    ) -> None:
        self.protocol = protocol
        self.compresslevel = compresslevel

    def encode(self, node: MeshNodeSpec) -> bytes:
        data = pickle.dumps(node, protocol=self.protocol)
        return gzip.compress(data, compresslevel=self.compresslevel)

    def decode(self, data: bytes) -> MeshNodeSpec:
        data = gzip.decompress(data)
        return pickle.loads(data)



import os
from asyncio import CancelledError

from rosy.types import DomainId

ALLOWED_EXCEPTIONS = (
    CancelledError,
    KeyboardInterrupt,
    SystemExit,
    GeneratorExit,
)

DEFAULT_DOMAIN_ID: DomainId = "default"


def get_domain_id(default: DomainId = DEFAULT_DOMAIN_ID) -> DomainId:
    return os.environ.get("ROSY_DOMAIN_ID", default)


def require(result, message: str = None) -> None:
    if not result:
        raise ValueError(message) if message else ValueError()



import asyncio
from argparse import ArgumentParser, Namespace
from typing import Iterable

from rosy.discovery.zeroconf import ZeroconfNodeDiscovery
from rosy.specs import MeshTopologySpec


def add_discovery_time_arg(parser: ArgumentParser) -> None:
    parser.add_argument(
        '--discovery-time',
        type=float,
        default=5,
        help='Node discovery time in seconds. Default: %(default)s'
    )


def add_log_arg(parser: ArgumentParser, default: str = 'WARNING') -> None:
    parser.add_argument(
        '--log',
        default=default,
        help='Log level; DEBUG, INFO, ERROR, etc. Default: %(default)s'
    )


async def get_mesh_topology(args: Namespace) -> MeshTopologySpec:
    async with ZeroconfNodeDiscovery(args.domain_id) as discovery:
        await asyncio.sleep(args.discovery_time)
        return discovery.topology


def print_args_and_kwargs(args: Iterable, kwargs: dict) -> None:
    if args:
        print('args:')
        for i, arg in enumerate(args):
            arg = arg_to_str(arg)
            print(f'  {i}: {arg}')

    if kwargs:
        print('kwargs:')
        for key, value in kwargs.items():
            value = arg_to_str(value)
            print(f'  {key}={value}')


def arg_to_str(arg) -> str:
    arg = repr(arg)

    if '\n' in arg:
        arg = f'```\n{arg}\n```'

    return arg



import asyncio
import logging

from rosy import build_node
from rosy.types import Topic


async def main():
    logging.basicConfig(level="WARNING")

    async with await build_node("topic_listener") as node:
        await node.listen("some-topic", callback)
        print("Listening...")
        await node.forever()


async def callback(topic: Topic, message: str, name: str = None):
    print(f'Received "{message} {name}" on topic={topic}')


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import logging

from rosy import build_node


async def main():
    logging.basicConfig(level="WARNING")

    async with await build_node(name="service_caller") as node:
        print("Calculating 2 * 2...")
        result = await node.call("multiply", 2, 2)
        print(f"Result: {result}")


if __name__ == "__main__":
    asyncio.run(main())



import logging

from rosy.node.topic.listenermanager import TopicListenerManager
from rosy.node.topic.types import TopicMessage
from rosy.utils import ALLOWED_EXCEPTIONS

logger = logging.getLogger(__name__)


class TopicMessageHandler:
    def __init__(
            self,
            listener_manager: TopicListenerManager,
    ):
        self.listener_manager = listener_manager

    async def handle_message(self, message: TopicMessage) -> None:
        callback = self.listener_manager.get_callback(message.topic)

        if not callback:
            logger.warning(
                f'Received message for topic={message.topic!r} '
                f'but no listener is registered.'
            )

        try:
            await callback(message.topic, *message.args, **message.kwargs)
        except ALLOWED_EXCEPTIONS:
            raise
        except Exception as e:
            logger.exception(
                f'Error calling callback={callback} '
                f'for topic={message.topic!r} '
                f'with args={message.args!r} '
                f'and kwargs={message.kwargs!r}',
                exc_info=e,
            )



import asyncio
import logging

from rosy import build_node


async def main():
    logging.basicConfig(level="WARNING")

    async with await build_node(name="topic_sender") as node:
        await node.send("some-topic", "hello", name="world")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import logging

from rosy import build_node
from rosy.types import Service


async def main():
    logging.basicConfig(level="WARNING")

    async with await build_node("service_provider") as node:
        await node.add_service("multiply", multiply)
        print("Started service...")
        await node.forever()


async def multiply(service: Service, a: int, b: int) -> int:
    return a * b


if __name__ == "__main__":
    asyncio.run(main())
